import logging

from typing import Callable
from dataclasses import dataclass
from collections import defaultdict
from almabtrieb import Almabtrieb

from .util import call_handler, HandlerInformation

logger = logging.getLogger(__name__)


@dataclass
class HandlerConfiguration:
    action: str
    activity_type: str
    func: Callable | None = None


class Handlers:
    def __init__(self):
        self.handler_map = defaultdict(lambda: defaultdict(list))

    @property
    def has_handlers(self):
        return len(self.handler_map) > 0

    def add_handler(self, config: HandlerConfiguration, func):
        self.handler_map[config.action][config.activity_type].append(
            HandlerInformation(func=func)
        )

    async def handle(
        self,
        data: dict,
        event_type: str,
        connection: Almabtrieb,
        actor_id: str | None = None,
        cow=None,
    ):
        activity = data.get("data", {}).get("raw", {})
        data_activity_type = activity.get("type")

        if actor_id is None:
            logger.warning("Skipping handlers due to missing actor_id")
            return

        for action in [event_type, "*"]:
            for activity_type in [data_activity_type, "*"]:
                handlers = self.handler_map[action][activity_type]
                for handler_info in handlers:
                    await call_handler(
                        handler_info, data, connection, actor_id=actor_id, cow=cow
                    )



import asyncio
from unittest.mock import AsyncMock

from .scheduler import HerdScheduler


async def test_empty_scheduler_exits():
    scheduler = HerdScheduler(entries=[], connection=AsyncMock())

    await scheduler.run()


async def test_scheduler_runs_with_task_group_exits():
    scheduler = HerdScheduler(entries=[], connection=AsyncMock())

    async with asyncio.TaskGroup() as tg:
        scheduler.create_task(tg)



from fast_depends import Depends
from typing import Annotated

from .common import PublishObject, PublishActivity  # noqa


def get_raw(data: dict) -> dict:
    return data.get("data", {}).get("raw")


def get_parsed(data: dict) -> dict:
    result = data.get("data", {}).get("parsed")
    if result is None:
        raise ValueError("No parsed data found")
    return result


RawData = Annotated[dict, Depends(get_raw)]
"""The raw data as received by cattle_grid"""

ParsedData = Annotated[dict, Depends(get_parsed)]
"""The parsed data as transformed by muck_out"""


def get_activity(parsed: ParsedData) -> dict:
    result = parsed.get("activity")
    if not result:
        raise ValueError("No activity found")
    return result


def get_embedded_object(parsed: ParsedData) -> dict:
    result = parsed.get("embeddedObject")
    if not result:
        raise ValueError("No embedded object found")
    return result


Activity = Annotated[dict, Depends(get_activity)]
"""The activity parsed by muck_out"""

EmbeddedObject = Annotated[dict, Depends(get_embedded_object)]
"""The embedded object in the activity as parsed by muck_out"""



from unittest.mock import AsyncMock
from almabtrieb.mqtt import MqttConnection

from .util import HandlerInformation, call_handler


async def test_call_handler():
    test_data = {"test": "data"}

    async def func(data: dict):
        assert data == test_data

    handler_info = HandlerInformation(func=func)

    connection = AsyncMock(MqttConnection)

    await call_handler(handler_info, test_data, connection, "actor_id")



"""Test documentation"""

from typing import Annotated, Awaitable, Callable
from fast_depends import Depends
from .common import Profile, PublishObject

try:
    from bovine.activitystreams import factories_for_actor_object
    from bovine.activitystreams.activity_factory import (
        ActivityFactory as BovineActivityFactory,  # type: ignore
    )
    from bovine.activitystreams.object_factory import (
        ObjectFactory as BovineObjectFactory,  # type: ignore
    )

    def get_activity_factory(profile: Profile) -> BovineActivityFactory:  # type: ignore
        activity_factory, _ = factories_for_actor_object(profile)
        return activity_factory  # type: ignore

    def get_object_factory(profile: Profile) -> BovineObjectFactory:  # type: ignore
        _, object_factory = factories_for_actor_object(profile)
        return object_factory  # type: ignore

except ImportError:

    class BovineActivityFactory: ...

    class BovineObjectFactory: ...

    def get_activity_factory() -> None:
        raise ImportError("bovine not installed")

    def get_object_factory() -> None:
        raise ImportError("bovine not installed")


ActivityFactory = Annotated[BovineActivityFactory, Depends(get_activity_factory)]  # type: ignore
"""The activity factory of type [bovine.activitystreams.activity_factory.ActivityFactory][]"""

ObjectFactory = Annotated[BovineObjectFactory, Depends(get_object_factory)]  # type: ignore
"""The object factory of type [bovine.activitystreams.object_factory.ObjectFactory][]"""


try:
    import markdown

    def get_markdown_poster(  # type: ignore
        object_factory: ObjectFactory, publish_object: PublishObject
    ):
        async def publish_markdown(message: str):
            content = markdown.markdown(message)
            note = object_factory.note(content=content).as_public().build()  # type: ignore
            await publish_object(note)

        return publish_markdown

except ImportError:

    def get_markdown_poster() -> None:
        raise ImportError("bovine not installed")


MarkdownPublisher = Callable[[str], Awaitable[None]]
"""Type of the markdown publisher"""

MarkdownPoster = Annotated[MarkdownPublisher, Depends(get_markdown_poster)]  # type: ignore
"""A function that takes a markdown string and posts it as the content of a note"""



import logging

from typing import Any, Callable, List

from dataclasses import dataclass, field
from fast_depends import inject

from cron_descriptor import get_description
from almabtrieb import Almabtrieb

from .types import Information
from .handlers import Handlers, HandlerConfiguration
from .profile import determine_profile_update

logger = logging.getLogger(__name__)


@dataclass
class CronEntry:
    """A cron entry"""

    crontab: str = field(metadata=dict(description="""The cron expression"""))

    func: Callable = field(metadata=dict(description="""The function to be called"""))


@dataclass
class RoboCowInternals:
    """Internal data for the cow"""

    profile: dict[str, Any] | None = field(
        default=None,
        metadata=dict(
            description="""The profile of the cow, aka as the actor object in ActivityPub"""
        ),
    )

    actor_id: str | None = field(
        default=None,
        metadata=dict(description="""Actor Id of the cow; loaded automatically"""),
    )

    handlers: Handlers = field(
        default_factory=Handlers,
        metadata=dict(
            description="""Handlers for incoming and outgoing messages, added through annotations"""
        ),
    )
    handler_configuration: List[HandlerConfiguration] = field(
        default_factory=list,
        metadata=dict(
            description="""Handler configurations, added through annotations"""
        ),
    )

    cron_entries: List[CronEntry] = field(
        default_factory=list,
        metadata=dict(description="""Cron entries, created through annotations"""),
    )

    startup_routine: Callable | None = None

    base_url: str | None = field(default=None)


@dataclass
class RoboCow:
    information: Information = field(
        metadata=dict(description="Information about the cow")
    )

    auto_follow: bool = field(
        default=True,
        metadata=dict(
            description="""Whether to automatically accept follow requests"""
        ),
    )

    skip_profile_update: bool = field(
        default=False,
        metadata=dict(
            description="When set to True the profile is not updated automatically. Useful when managing a cow from multiple scripts."
        ),
    )

    internals: RoboCowInternals = field(
        default_factory=RoboCowInternals,
        metadata=dict(description="Internal data for the cow"),
    )

    @staticmethod
    def create(**kwargs):
        """Creates a new cow, by creating a new [Information][roboherd.cow.types.Information].

        ```python
        >>> RoboCow.create(name="my cow")
        RoboCow(information=Information(type='Service', handle=None, name='my cow', ...

        ```

        The parameter `description_md` is transformed from markdown to html and then
        assigned to `description`.

        ```python
        >> cow = RoboCow.create(description_md="__bold__")
        >> cow.information.description
        '<p><strong>bold</strong></p>'

        ```
        """

        if "description_md" in kwargs:
            import markdown

            kwargs["description"] = markdown.markdown(kwargs["description_md"])
            del kwargs["description_md"]

        information = Information.model_validate(kwargs)
        return RoboCow(information=information)

    def action(self, action: str = "*", activity_type: str = "*"):
        """Adds a handler for an event. Use "*" as a wildcard.

        Usage:

        ```python
        cow = Robocow(information=Information(handle="example"))

        @cow.action(action="outgoing", activity_type="Follow")
        async def handle_outgoing_follow(data):
            ...
        ```
        """

        config = HandlerConfiguration(
            action=action,
            activity_type=activity_type,
        )

        def inner(func):
            config.func = func
            self.internals.handlers.add_handler(config, func)
            self.internals.handler_configuration.append(config)
            return func

        return inner

    def cron(self, crontab):
        def inner(func):
            self.internals.cron_entries.append(CronEntry(crontab, func))

            return func

        return inner

    def incoming(self, func):
        """Adds a handler for an incoming message. Usage:

        ```python
        cow = Robocow("example")

        @cow.incoming
        async def handle_incoming(data):
            ...
        ```
        """
        config = HandlerConfiguration(
            action="incoming",
            activity_type="*",
        )
        self.internals.handlers.add_handler(config, func)
        return func

    def incoming_create(self, func):
        """Adds a handler for an incoming activity if the
        activity is of type_create

        ```python
        cow = Robocow("example")

        @cow.incoming_create
        async def handle_incoming(data):
            ...
        ```
        """
        config = HandlerConfiguration(
            action="incoming", activity_type="Create", func=func
        )
        self.internals.handler_configuration.append(config)
        self.internals.handlers.add_handler(config, func)
        return func

    def startup(self, func):
        """Adds a startup routine to be run when the cow is started."""

        self.internals.startup_routine = func

    async def run_startup(self, connection: Almabtrieb):
        """Runs when the cow is birthed"""

        if self.internals.profile is None:
            if not self.internals.actor_id:
                raise ValueError("Actor ID is not set")
            result = await connection.fetch(
                self.internals.actor_id, self.internals.actor_id
            )
            if not result.data:
                raise ValueError("Could not retrieve profile")
            self.internals.profile = result.data

        if self.internals.cron_entries:
            frequency = ", ".join(
                get_description(entry.crontab) for entry in self.internals.cron_entries
            )
            self.information.frequency = frequency

        if not self.skip_profile_update:
            await self._run_profile_update(connection)

        if self.internals.startup_routine:
            await inject(self.internals.startup_routine)(
                cow=self,  # type:ignore
                connection=connection,  # type:ignore
                actor_id=self.internals.actor_id,  # type:ignore
            )  # type:ignore

    async def _run_profile_update(self, connection: Almabtrieb):
        if self.internals.profile is None:
            raise ValueError("Profile is not set")

        update = determine_profile_update(self.information, self.internals.profile)

        if update:
            logger.info("Updating profile for %s", self.information.handle)

            await connection.trigger("update_actor", update)



from pydantic import BaseModel, Field

from .const import default_icon


class MetaInformation(BaseModel):
    """Meta Information about the bot. This includes
    information such as the author and the source repository"""

    source: str | None = Field(
        default=None,
        examples=["https://forge.example/repo"],
        description="The source repository",
    )

    author: str | None = Field(
        default=None,
        examples=["acct:author@domain.example"],
        description="The author, often a Fediverse handle",
    )


class Information(BaseModel):
    """Information about the cow"""

    type: str = Field(
        default="Service",
        examples=["Service"],
        description="ActivityPub type of the actor.",
    )

    handle: str | None = Field(
        default=None,
        examples=["moocow"],
        description="Used as the handle in `acct:handle@domain.example`",
    )

    name: str | None = Field(
        default=None,
        examples=["The mooing cow 🐮"],
        description="The display name of the cow",
    )

    description: str | None = Field(
        default=None,
        examples=[
            "I'm a cow that moos.",
            """<p>An example bot to illustrate Roboherd</p><p>For more information on RoboHerd, see <a href="https://codeberg.org/bovine/roboherd">its repository</a>.</p>""",
        ],
        description="The description of the cow, used as summary of the actor",
    )

    icon: dict = Field(
        default=default_icon,
        description="The profile image",
    )

    frequency: str | None = Field(
        default=None,
        examples=["daily"],
        description="Frequency of posting. Is set automatically if cron expressions are used.",
    )

    meta_information: MetaInformation = Field(
        default=MetaInformation(),
        description="Meta information about the cow, such as the source repository",
    )



from pydantic import BaseModel, Field

from dataclasses import dataclass, field

from roboherd.cow import RoboCow
from .load import load_cow


class ConfigOverrides(BaseModel):
    """Values used in `roboherd.toml` to overide the default
    values in the imported cow. This class is meant as a
    reference, and not meant to be directly used."""

    name: str | None = Field(
        default=None, description="set to override the name", examples=["New name"]
    )
    handle: str | None = Field(
        default=None, description="set to override the handle", examples=["new-handle"]
    )
    base_url: str | None = Field(
        default=None,
        description="set to override the base url",
        examples=["https://other.example"],
    )
    skip_profile_update: bool | None = Field(
        default=None, description="set to skip updating the profile", examples=[True]
    )


@dataclass
class CowConfig:
    name: str = field(metadata={"description": "Name of the cow, must be unique"})
    module: str
    attribute: str
    config: dict

    @staticmethod
    def from_name_and_dict(name, cow: dict) -> "CowConfig":
        module, attribute = cow["bot"].split(":")

        return CowConfig(name=name, module=module, attribute=attribute, config=cow)

    def load(self) -> RoboCow:
        cow = load_cow(self.module, self.attribute)

        overrides = ConfigOverrides(**self.config)

        for value in ["name", "handle", "base_url"]:
            if getattr(overrides, value):
                setattr(cow.information, value, getattr(overrides, value))

        if overrides.skip_profile_update:
            cow.skip_profile_update = overrides.skip_profile_update

        return cow

    def __hash__(self):
        return hash(self.name)


@dataclass
class HerdConfig:
    cows: list[CowConfig] = field(default_factory=list)

    def for_name(self, name: str) -> CowConfig | None:
        for cow in self.cows:
            if cow.name == name:
                return cow
        return None

    @property
    def names(self) -> set[str]:
        return {cow.name for cow in self.cows}

    @staticmethod
    def from_settings(settings):
        cows = [
            CowConfig.from_name_and_dict(name, config)
            for name, config in settings.cow.items()
        ]

        return HerdConfig(cows=cows)

    def load_herd(self) -> list[RoboCow]:
        return [cow.load() for cow in self.cows]



import tomli_w


def create_config(
    name: str, password: str, domain: str = "dev.bovine.social"
) -> dict[str, bool | str]:
    """

    ```
    >>> create_config("alice", "pass")
    {'base_url': 'https://dev.bovine.social',
        'connection_string': 'wss://alice:pass@dev.bovine.social/mooqtt',
        'echo': False}

    ```
    """

    return {
        "base_url": f"https://{domain}",
        "connection_string": f"wss://{name}:{password}@{domain}/mooqtt",
        "echo": False,
    }


async def register(config_file: str, name: str, password: str, fediverse: str):
    import aiohttp

    async with aiohttp.ClientSession() as session:
        result = await session.post(
            "https://dev.bovine.social/register",
            json={
                "name": name,
                "password": password,
                "fediverse": fediverse,
            },
        )

        assert result.status == 201, "Something went wrong"

    with open(config_file, "wb") as f:
        tomli_w.dump(create_config(name, password), f)



import logging

from dataclasses import dataclass, field

from almabtrieb.model import ActorInformation

from roboherd.cow import RoboCow

from .config import HerdConfig, CowConfig

logger = logging.getLogger(__name__)


@dataclass
class HerdManager:
    prefix: str = "bot:"
    herd_config: HerdConfig = field(default_factory=HerdConfig)

    @staticmethod
    def from_settings(settings):
        return HerdManager(herd_config=HerdConfig.from_settings(settings))

    def existing_cows(self, actors: list[ActorInformation]) -> list[RoboCow]:
        existing_cows = []

        for info in actors:
            if info.name.startswith(self.prefix):
                cow_name = info.name.removeprefix(self.prefix)
                cow_config = self.herd_config.for_name(cow_name)
                if cow_config:
                    cow = cow_config.load()
                    cow.internals.actor_id = info.id
                    existing_cows.append(cow)

        return existing_cows

    def cows_to_create(self, existing_actors: list[ActorInformation]) -> set[CowConfig]:
        existing_names = {
            actor.name.removeprefix(self.prefix)
            for actor in existing_actors
            if actor.name.startswith(self.prefix)
        }
        names_to_create = self.herd_config.names - existing_names

        cows = {self.herd_config.for_name(name) for name in names_to_create}

        return {cow for cow in cows if cow}



from roboherd import RoboCow, PublishObject, ObjectFactory

from .meta import meta_information

bot = RoboCow.create(
    handle="rooster",
    name="The crowing rooster 🐓",
    description="I'm a rooster that crows at a set frequency.",
    meta_information=meta_information,
)


@bot.cron("42 * * * *")
async def crow(publisher: PublishObject, object_factory: ObjectFactory):
    await publisher(
        object_factory.note(content="cock-a-doodle-doo").as_public().build()  # type: ignore
    )



import copy
import importlib
from importlib import import_module

from roboherd.cow import RoboCow


def load_cow(module_name: str, attribute: str) -> RoboCow:
    """Loads a cow from module name and attribute"""
    module = import_module(module_name)
    importlib.reload(module)

    cow = getattr(module, attribute)

    return copy.deepcopy(cow)



from roboherd import RoboCow

from .meta import meta_information

bot = RoboCow.create(
    name="/dev/null",
    description="""I don't do anything.""",
    handle="devnull",
    meta_information=meta_information,
)



import pytest
from unittest.mock import AsyncMock

from . import RoboCow
from .types import Information


def test_cron():
    info = Information(handle="testcow")
    cow = RoboCow(information=info)

    @cow.cron("* * * * *")
    async def test_func():
        pass

    assert len(cow.internals.cron_entries) == 1


async def test_startup():
    info = Information(handle="testcow")
    cow = RoboCow(information=info)
    cow.internals.profile = {"id": "http://host.test/actor/cow"}
    mock = AsyncMock()
    connection = AsyncMock()

    cow.startup(mock)

    await cow.run_startup(connection=connection)

    mock.assert_called_once()
    connection.trigger.assert_awaited_once()


async def test_skip_startup():
    info = Information(handle="testcow")
    cow = RoboCow(information=info, skip_profile_update=True)
    cow.internals.profile = {"id": "http://host.test/actor/cow"}

    connection = AsyncMock()
    await cow.run_startup(connection=connection)

    connection.trigger.assert_not_awaited()


@pytest.mark.skipif(
    not pytest.importorskip("markdown"), reason="markdown not installed"
)
def test_create():
    cow = RoboCow.create(handle="testcow", description_md="__bold__")

    assert cow.information.handle == "testcow"
    assert cow.information.description == "<p><strong>bold</strong></p>"



from roboherd.cow import RoboCow
from roboherd.examples.moocow import moocow  # noqa


def create_actor_message_for_cow(cow: RoboCow, base_url):
    """
    ```pycon
    >>> create_actor_message_for_cow(moocow, "http://domain.example/")
    {'baseUrl': 'http://domain.example/',
        'preferredUsername': 'moocow',
        'automaticallyAcceptFollowers': True,
        'profile': {'type': 'Service'}}

    ```
    """
    return {
        "baseUrl": base_url,
        "preferredUsername": cow.information.handle,
        "automaticallyAcceptFollowers": cow.auto_follow,
        "profile": {"type": "Service"},
    }



import pytest
from dynaconf import Dynaconf

from almabtrieb.model import ActorInformation


from . import HerdManager


@pytest.fixture
def test_config():
    config = Dynaconf()
    config.update(
        {
            "cow": {
                "moocow": {
                    "bot": "roboherd.examples.moocow:moocow",
                },
                "rooster": {
                    "bot": "roboherd.examples.rooster:bot",
                },
            },
        }
    )  # type: ignore
    return config


@pytest.mark.parametrize(
    "actor_info,result_lenth",
    [
        ([], 2),
        ([ActorInformation(name="bot:moocow", id="http://host.test/actor")], 1),
        ([ActorInformation(name="moocow", id="http://host.test/actor")], 2),
        ([ActorInformation(name="bot:other", id="http://host.test/actor")], 2),
    ],
)
def test_cows_to_create(test_config, actor_info, result_lenth):
    manager = HerdManager.from_settings(test_config)

    result = manager.cows_to_create(actor_info)

    assert len(result) == result_lenth



import asyncio
import logging
import os

import click
import dynaconf

from roboherd.herd import RoboHerd
from roboherd.herd.manager import HerdManager
from roboherd.util import create_connection
from roboherd.register import register as run_register
from roboherd.validators import validators
from roboherd.version import __version__

logging.basicConfig(level=logging.INFO)
logging.captureWarnings(True)


@click.group(invoke_without_command=True)
@click.option(
    "--connection_string",
    default=None,
    help="Connection string to the websocket mqtt broker",
)
@click.option(
    "--base_url",
    default=None,
    help="Base url to create cows with",
)
@click.option("--config_file", default="roboherd.toml", help="Configuration file")
@click.option(
    "--version", is_flag=True, default=False, help="display version then exit"
)
@click.pass_context
def main(
    ctx: click.Context,
    connection_string: str,
    base_url: str,
    config_file: str,
    version: bool,
):
    """Configuration is usually loaded from the config_file. These options can be overwritten by passing as a command line argument."""
    if version:
        print(f"roboherd version: {__version__}")
        exit(0)

    settings = dynaconf.Dynaconf(
        settings_files=[config_file],
        envvar_prefix="ROBOHERD",
        validators=validators,
    )
    ctx.ensure_object(dict)

    ctx.obj["config_file"] = config_file
    ctx.obj["settings"] = settings

    if connection_string:
        ctx.obj["connection_string"] = connection_string
    else:
        ctx.obj["connection_string"] = settings.connection_string  # type: ignore

    if base_url:
        ctx.obj["base_url"] = base_url
    else:
        ctx.obj["base_url"] = settings.base_url  # type: ignore

    print("Please specify a command")


@main.command()
@click.option("--fail", is_flag=True, default=False, help="Fail if actors do not exist")
@click.pass_context
def check(ctx: click.Context, fail: bool):
    """Checks that the connection is configured correctly"""

    create_connection(ctx)

    herd = RoboHerd(base_url=ctx.obj["base_url"])

    settings = ctx.obj["settings"]

    if settings.get("cow"):
        herd.manager = HerdManager.from_settings(settings)
        asyncio.run(herd.check(ctx.obj["connection"], raise_if_cows_to_create=fail))
    else:
        click.echo("No cows specified")
        exit(1)


@main.command()
@click.pass_context
def run(ctx: click.Context):
    """Runs the roboherd by connecting to the server."""

    create_connection(ctx)

    herd = RoboHerd(base_url=ctx.obj["base_url"])

    settings = ctx.obj["settings"]

    if settings.get("cow"):
        herd.manager = HerdManager.from_settings(settings)
        asyncio.run(herd.run(ctx.obj["connection"]))
    else:
        click.echo("No cows specified")
        exit(1)


@main.command()
@click.pass_context
def watch(ctx):
    """Watches the file the module is in for changes and then restarts roboherd.

    Note: Options for roboherd are currently ignored (FIXME)."""
    import watchfiles

    watchfiles.run_process("roboherd", target="roboherd run")


@main.command()
@click.pass_context
@click.option("--name", help="Name for the account to be created", prompt=True)
@click.option(
    "--password",
    help="Password for the account to be created",
    hide_input=True,
    prompt=True,
)
@click.option("--fediverse", help="Fediverse handle", prompt=True)
def register(ctx: click.Context, name: str, password: str, fediverse: str):
    """Registers a new account on dev.bovine.social. All three options are required. If not provided, you will be prompted for them."""

    if os.path.exists(ctx.obj["config_file"]):
        click.echo("Config file already exists")
        exit(1)

    if len(password) < 6:
        click.echo("Password should have at least 6 characters")
        exit(1)

    asyncio.run(run_register(ctx.obj["config_file"], name, password, fediverse))


if __name__ == "__main__":
    main()



import click
import logging

from urllib.parse import urlparse
from almabtrieb import Almabtrieb


logger = logging.getLogger(__name__)


def parse_connection_string(connection_string: str) -> dict[str, str | int | None]:
    """
    Parse a connection string into a dictionary of connection parameters.

    ```pycon
    >>> parse_connection_string("ws://user:pass@host/ws")
    {'host': 'host',
        'port': 80,
        'username': 'user',
        'password': 'pass',
        'websocket_path': '/ws'}

    >>> parse_connection_string("wss://user:pass@host/ws")
    {'host': 'host',
        'port': 443,
        'username': 'user',
        'password': 'pass',
        'websocket_path': '/ws'}

    ```
    """

    parsed = urlparse(connection_string)

    default_port = 80 if parsed.scheme == "ws" else 443

    return {
        "host": parsed.hostname,
        "port": parsed.port or default_port,
        "username": parsed.username,
        "password": parsed.password,
        "websocket_path": parsed.path,
    }


def create_connection(ctx: click.Context):
    connection_string = ctx.obj["connection_string"]
    base_url = ctx.obj["base_url"]

    if not connection_string:
        click.echo("ERROR: No connection string provided")
        click.echo(
            "either provide one through --connection_string or set it in your configuration file"
        )
        exit(1)

    if not base_url:
        click.echo("ERROR: No base url for cows provided")
        click.echo(
            "either provide one through --base_url or set it in your configuration file"
        )
        exit(1)

    ctx.obj["connection"] = Almabtrieb.from_connection_string(
        connection_string, echo=ctx.obj["settings"].get("echo", False)
    )



import random
from urllib.parse import urlparse

from almabtrieb import Almabtrieb

from roboherd import RoboCow, PublishObject, ObjectFactory
from .meta import meta_information


def hostname(actor_id):
    return urlparse(actor_id).netloc


bot = RoboCow.create(
    handle="even",
    description="""<p>I'm a bot by Helge. I post a random number every hour. When posting an even number, I change my Fediverse handle to even. When posting an odd one, I use odd.</p>

<p>I also update my name. I'm not sure how you should display my handle with your Fediverse application. Please write a FEP explaining it.</p>""",
    meta_information=meta_information,
)


@bot.startup
async def startup(connection: Almabtrieb, actor_id: str):
    await connection.trigger(
        "update_actor",
        {
            "actor": actor_id,
            "actions": [
                {
                    "action": "add_identifier",
                    "identifier": f"acct:odd@{hostname(actor_id)}",
                    "primary": False,
                }
            ],
        },
    )


@bot.cron("* * * * *")
async def post_number(
    connection: Almabtrieb,
    publisher: PublishObject,
    factory: ObjectFactory,
    actor_id: str,
):
    number = random.randint(0, 1000)

    note = factory.note(content=f"Number: {number}").as_public().build()  # type: ignore
    await publisher(note)

    handle = "even" if number % 2 == 0 else "odd"

    await connection.trigger(
        "update_actor",
        {
            "actor": actor_id,
            "actions": [
                {
                    "action": "update_identifier",
                    "identifier": f"acct:{handle}@{hostname(actor_id)}",
                    "primary": True,
                }
            ],
            "profile": {"name": f"Posted an {handle} number"},
        },
    )



from dynaconf import Dynaconf

from .validators import validators


def test_validators():
    settings = Dynaconf(validators=validators)

    assert settings.base_url is None
    assert settings.connection_string is None



from fast_depends import Depends
from typing import Annotated, Any, Callable, Awaitable

from almabtrieb import Almabtrieb


from roboherd.cow import RoboCow


def get_profile(cow: RoboCow) -> dict[str, Any]:
    if cow.internals.profile is None:
        raise ValueError("Cow has no profile")
    return cow.internals.profile


Profile = Annotated[dict[str, Any], Depends(get_profile)]
"""The profile of the cow"""


Publisher = Callable[[dict[str, Any]], Awaitable[None]]
"""Type returned by the publishing functions"""


def construct_publish_object(connection: Almabtrieb, actor_id: str) -> Publisher:
    async def publish(data: dict):
        await connection.trigger("publish_object", {"actor": actor_id, "data": data})

    return publish


def construct_publish_activity(connection: Almabtrieb, actor_id: str) -> Publisher:
    async def publish(data: dict):
        await connection.trigger("publish_activity", {"actor": actor_id, "data": data})

    return publish


PublishObject = Annotated[Publisher, Depends(construct_publish_object)]
"""Allows one to publish an object as the actor. Assumes cattle_grid has the extension `simple_object_storage` or equivalent"""


PublishActivity = Annotated[Publisher, Depends(construct_publish_activity)]
"""Allows one to publish an activity as the actor. Assumes cattle_grid has the extension `simple_object_storage` or equivalent"""



from urllib.parse import urlparse

try:
    from bovine.activitystreams.utils import as_list
except ImportError:

    def as_list(value):
        if isinstance(value, list):
            return value

        return [value]


from .types import Information


def profile_part_needs_update(information: Information, profile: dict) -> bool:
    if information.name != profile.get("name"):
        return True

    if information.description != profile.get("summary"):
        return True

    if information.type != profile.get("type"):
        return True

    if information.icon != profile.get("icon"):
        return True

    return False


def key_index_from_attachment(attachments: list[dict], key: str) -> int | None:
    for idx, attachment in enumerate(attachments):
        if attachment is None:
            continue
        if attachment.get("type") == "PropertyValue" and attachment.get("name") == key:
            return idx
    return None


def determine_action_for_key_and_value(
    attachments: list[dict], key: str, value: str | None
) -> dict | None:
    idx = key_index_from_attachment(attachments, key)
    if idx is None:
        if value:
            return {
                "action": "update_property_value",
                "key": key,
                "value": value,
            }
        return None

    if value is None:
        return {
            "action": "remove_property_value",
            "key": key,
        }
    current_value = attachments[idx].get("value")
    if value != current_value:
        return {
            "action": "update_property_value",
            "key": key,
            "value": value,
        }

    return None


def determine_actions(information: Information, profile: dict) -> list[dict] | None:
    attachments = as_list(profile.get("attachment", []))
    meta_information = information.meta_information

    actions = [
        determine_action_for_key_and_value(
            attachments, "Author", meta_information.author
        ),
        determine_action_for_key_and_value(
            attachments, "Source", meta_information.source
        ),
        determine_action_for_key_and_value(
            attachments, "Frequency", information.frequency
        ),
    ]

    if information.handle and profile.get("preferredUsername") is None:
        actions.append(
            {
                "action": "add_identifier",
                "identifier": "acct:"
                + information.handle
                + "@"
                + str(urlparse(profile.get("id")).netloc),
                "primary": True,
            }
        )

    actions = [x for x in actions if x is not None]

    if len(actions) == 0:
        return None
    return actions


def determine_profile_update(information: Information, profile: dict) -> dict | None:
    """Returns the update for the profile"""

    update = {"actor": profile.get("id")}

    if profile_part_needs_update(information, profile):
        update["profile"] = {
            "type": information.type,
            "name": information.name,
            "summary": information.description,
            "icon": information.icon,
        }

    actions = determine_actions(information, profile)

    if actions:
        update["actions"] = actions

    if len(update) == 1:
        return None

    return update



import pytest

from unittest.mock import AsyncMock

from almabtrieb.mqtt import MqttConnection

from .handlers import Handlers, HandlerConfiguration


@pytest.mark.parametrize(
    "action,activity_type",
    [
        ("incoming", "AnimalSound"),
        ("incoming", "*"),
        ("*", "AnimalSound"),
        ("*", "*"),
    ],
)
async def test_handlers_should_run(action, activity_type):
    handlers = Handlers()
    connection = AsyncMock(MqttConnection)

    mock = AsyncMock()

    handlers.add_handler(
        HandlerConfiguration(action=action, activity_type=activity_type), mock
    )
    await handlers.handle(
        {"data": {"raw": {"type": "AnimalSound"}}},
        "incoming",
        connection,
        actor_id="actor_id",
    )

    mock.assert_awaited_once()


@pytest.mark.parametrize(
    "action,activity_type",
    [
        ("outgoing", "AnimalSound"),
        ("outgoing", "*"),
        ("*", "Create"),
        ("incoming", "Create"),
    ],
)
async def test_handlers_should_nod_run(action, activity_type):
    handlers = Handlers()
    connection = AsyncMock(MqttConnection)

    mock = AsyncMock()

    handlers.add_handler(
        HandlerConfiguration(action=action, activity_type=activity_type), mock
    )

    await handlers.handle(
        {"data": {"raw": {"type": "AnimalSound"}}},
        "incoming",
        connection,
        actor_id="actor_id",
    )

    mock.assert_not_awaited()


def test_has_handlers():
    handlers = Handlers()

    assert not handlers.has_handlers

    handlers.add_handler(
        HandlerConfiguration(action="outgoing", activity_type="Create"), AsyncMock()
    )

    assert handlers.has_handlers



"""Main roboherd module. Should contain all necessary elements to build a robocow"""

from .cow import RoboCow
from .annotations import (
    PublishObject,
    PublishActivity,
    RawData,
    ParsedData,
    Activity,
    EmbeddedObject,
)
from .annotations.bovine import MarkdownPoster, ObjectFactory, ActivityFactory

__all__ = [
    "RoboCow",
    "RawData",
    "ParsedData",
    "Activity",
    "EmbeddedObject",
    "MarkdownPoster",
    "ActivityFactory",
    "ObjectFactory",
    "PublishObject",
    "PublishActivity",
]



from fast_depends import inject

from dataclasses import dataclass
from typing import Callable, Dict, Awaitable

from almabtrieb import Almabtrieb


@dataclass
class HandlerInformation:
    func: Callable[[Dict], Awaitable[None]]


async def call_handler(
    handler_info: HandlerInformation,
    data: dict,
    connection: Almabtrieb,
    actor_id: str,
    cow=None,
):
    return await inject(handler_info.func)(
        data=data,  # type: ignore
        connection=connection,
        actor_id=actor_id,
        cow=cow,
    )



from roboherd.cow import RoboCow

from .load import load_cow


def test_load_cow():
    cow = load_cow("roboherd.examples.moocow", "moocow")
    assert isinstance(cow, RoboCow)


def test_load_cow_can_overwrite_variables():
    one = load_cow("roboherd.examples.moocow", "moocow")

    one.information.name = "A cow eating watermelons"

    two = load_cow("roboherd.examples.moocow", "moocow")

    assert two.information.name == "The mooing cow 🐮"
    assert one.information.name == "A cow eating watermelons"



import asyncio
import logging
from dataclasses import dataclass
from typing import List

from almabtrieb import Almabtrieb

from roboherd.cow import RoboCow


logger = logging.getLogger(__name__)


@dataclass
class HerdProcessor:
    connection: Almabtrieb
    incoming_handlers: List[RoboCow]

    def create_tasks(self, task_group: asyncio.TaskGroup):
        tasks = []
        if len(self.incoming_handlers) > 0:
            tasks.append(task_group.create_task(self.process_incoming(self.connection)))

        return tasks

    async def process_incoming(self, connection):
        actor_id_to_cow_map = {}
        for cow in self.incoming_handlers:
            actor_id_to_cow_map[cow.internals.actor_id] = cow

        logger.info("Incoming processing started for %s cows", len(actor_id_to_cow_map))

        async for msg in connection.incoming():
            actor_id = msg["actor"]

            cow = actor_id_to_cow_map.get(actor_id)
            if cow:
                await cow.internals.handlers.handle(
                    msg, "incoming", connection, actor_id, cow=cow
                )

        logger.warning("Process incoming ended")



import pytest

from dynaconf import Dynaconf

from roboherd.cow import RoboCow

from .config import CowConfig, HerdConfig


@pytest.fixture
def test_config():
    config = Dynaconf()
    config.update(
        {
            "cow": {
                "moocow": {
                    "bot": "moocow:bot",
                },
                "rooster": {
                    "bot": "rooster:bot",
                },
            },
        }
    )  # type: ignore
    return config


def test_from_name_and_dict():
    name = "cow"
    value = {
        "bot": "module:attribute",
    }

    cow = CowConfig.from_name_and_dict(name, value)

    assert cow.name == name
    assert cow.module == "module"
    assert cow.attribute == "attribute"


def test_from_name_and_dict_with_new_name():
    name = "cow"
    value = {
        "bot": "roboherd.examples.moocow:moocow",
        "handle": "new_handle",
        "name": "new name",
        "skip_profile_update": True,
    }

    config = CowConfig.from_name_and_dict(name, value)

    cow = config.load()

    assert cow.information.name == "new name"
    assert cow.information.handle == "new_handle"
    assert cow.skip_profile_update is True


def test_load_config(test_config):
    herd = HerdConfig.from_settings(test_config)

    assert len(herd.cows) == 2

    moocow = herd.for_name("moocow")
    assert moocow
    assert moocow.name == "moocow"


def test_load_from_cow_config():
    name = "cow"
    value = {
        "bot": "roboherd.examples.moocow:moocow",
    }

    cow = CowConfig.from_name_and_dict(name, value)

    assert isinstance(cow.load(), RoboCow)


def test_names(test_config):
    herd = HerdConfig.from_settings(test_config)

    assert herd.names == {"moocow", "rooster"}



import pytest

from .types import Information, MetaInformation
from .const import default_icon
from .profile import determine_profile_update


@pytest.mark.parametrize(
    "info_params, profile",
    [
        ({}, {"type": "Service"}),
        ({"name": "name"}, {"type": "Service", "name": "name"}),
        ({"description": "description"}, {"type": "Service", "summary": "description"}),
    ],
)
def test_determine_profile_update_no_update(info_params, profile):
    profile["icon"] = default_icon
    info = Information(**info_params)

    assert determine_profile_update(info, profile) is None


def test_determine_profile_update():
    info = Information(name="name", description="description")
    profile = {"id": "http://host.test/actor/1"}

    result = determine_profile_update(info, profile)

    assert result == {
        "actor": "http://host.test/actor/1",
        "profile": {
            "type": "Service",
            "name": "name",
            "summary": "description",
            "icon": default_icon,
        },
    }


def test_determine_profile_update_author():
    info = Information(meta_information=MetaInformation(author="acct:author@host.test"))
    profile = {
        "id": "http://host.test/actor/1",
        "type": "Service",
        "icon": default_icon,
    }

    result = determine_profile_update(info, profile)

    assert result == {
        "actor": "http://host.test/actor/1",
        "actions": [
            {
                "action": "update_property_value",
                "key": "Author",
                "value": "acct:author@host.test",
            }
        ],
    }


def test_profile_update_for_new_preferredUsername():
    info = Information(handle="handle")
    profile = {
        "id": "http://host.test/actor/1",
        "type": "Service",
        "icon": default_icon,
    }

    result = determine_profile_update(info, profile)

    assert result == {
        "actor": "http://host.test/actor/1",
        "actions": [
            {
                "action": "add_identifier",
                "identifier": "acct:handle@host.test",
                "primary": True,
            }
        ],
    }



import logging

from roboherd import RoboCow, EmbeddedObject, PublishObject
from .meta import meta_information

logger = logging.getLogger(__name__)

moocow = RoboCow.create(
    handle="moocow",
    name="The mooing cow 🐮",
    description="""I'm a cow that likes to moo.

I also serve as an example for the RoboHerd python tool.
See <a href="https://codeberg.org/helge/roboherd">codeberg.org</a>.""",
    meta_information=meta_information,
)


@moocow.incoming_create
async def on_incoming_create(
    obj: EmbeddedObject, publisher: PublishObject, actor_id: str
):
    recipient = obj.get("attributedTo")

    logger.info("Replying to %s", recipient)

    obj = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "type": "Note",
        "attributedTo": actor_id,
        "to": [recipient],
        "cc": ["https://www.w3.org/ns/activitystreams#Public"],
        "content": "moo",
        "inReplyTo": obj.get("id"),
    }

    await publisher(obj)



import asyncio
import logging
from dataclasses import dataclass
from typing import List, Tuple

from apscheduler.schedulers.asyncio import AsyncIOScheduler
from apscheduler.triggers.cron import CronTrigger

from fast_depends import inject

from almabtrieb import Almabtrieb

from roboherd.cow import RoboCow, CronEntry

logger = logging.getLogger(__name__)


@dataclass
class HerdScheduler:
    entries: List[Tuple[RoboCow, CronEntry]]
    connection: Almabtrieb
    scheduler: AsyncIOScheduler | None = None
    task: asyncio.Task | None = None

    def create_task(self, task_group: asyncio.TaskGroup):
        if len(self.entries) == 0:
            logger.info("No tasks to schedule")
            return
        if self.task:
            raise Exception("Task already running")

        self.task = task_group.create_task(self.run())

    async def run(self):
        if len(self.entries) == 0:
            return

        if self.scheduler:
            raise Exception("Scheduler already exists")

        self.scheduler = AsyncIOScheduler()

        for cow, entry in self.entries:
            trigger = CronTrigger.from_crontab(entry.crontab)
            self.scheduler.add_job(
                inject(entry.func),
                trigger=trigger,
                kwargs={
                    "actor_id": cow.internals.actor_id,
                    "connection": self.connection,
                    "cow": cow,
                },
            )

        self.scheduler.start()

        while True:
            await asyncio.sleep(60 * 60)

    async def stop(self):
        logger.warning("Stopping scheduler")
        if self.scheduler:
            self.scheduler.shutdown()
        if self.task:
            self.task.cancel()



import json

from roboherd import RoboCow, RawData, PublishObject, ObjectFactory

from .meta import meta_information


def reply_content(raw: dict) -> str:
    """Formats and escapes the JSON data:

    ```pycon
    >>> reply_content({"html": "<b>bold</b>"})
    '<pre><code>{\\n  "html": "&lt;b&gt;bold&lt;/b&gt;"\\n}</code></re>'

    ```
    """
    json_formatted = (
        json.dumps(raw, indent=2)
        .replace("&", "&amp;")
        .replace("<", "&lt;")
        .replace(">", "&gt;")
    )
    return f"<pre><code>{json_formatted}</code></re>"


try:
    bot = RoboCow.create(
        handle="jsonecho",
        name="JSON Echo {}",
        description_md="""I'm a silly bot that replies to
    you with the JSON as received through a HTTP
    post request by [cattle_grid](https://bovine.codeberg.page/cattle_grid/).""",
        meta_information=meta_information,
    )

    @bot.incoming_create
    async def create(
        raw: RawData, publish_object: PublishObject, object_factory: ObjectFactory
    ):
        note = (
            object_factory.reply(  # type: ignore
                raw.get("object"),
                content=reply_content(raw),
            )
            .as_public()
            .build()
        )

        await publish_object(note)
except ImportError:
    ...



import pytest

from roboherd.examples.moocow import moocow
from roboherd.examples.rooster import bot
from roboherd.examples.dev_null import bot as dev_null

from roboherd.cow import CronEntry

from . import RoboHerd, RoboCow


@pytest.mark.parametrize("cow, length", [(moocow, 0), (bot, 1)])
def test_cron_entries(cow: RoboCow, length: int):
    manager = RoboHerd(cows=[cow])

    assert len(manager.cron_entries()) == length


def test_cron_entries_result():
    manager = RoboHerd(cows=[bot])

    result = manager.cron_entries()[0]

    assert result[0] == bot
    assert isinstance(result[1], CronEntry)


def test_incoming_handlers_no_result():
    manager = RoboHerd(cows=[dev_null])

    assert len(manager.incoming_handlers()) == 0


def test_incoming_handlers_result():
    manager = RoboHerd(cows=[moocow])

    assert len(manager.incoming_handlers()) == 1



from roboherd import RoboCow, MarkdownPoster

# from roboherd import PublishObject, ObjectFactory

from .meta import meta_information

bot = RoboCow.create(
    name="The scare crow 👩‍🌾",
    description="""On startup I scare crows""",
    handle="scarecrow",
    meta_information=meta_information,
)


# @bot.startup
# async def startup(publish_object: PublishObject, object_factory: ObjectFactory):
#     note = object_factory.note(content="Booo! 🐦").as_public().build()  # type: ignore
#     await publish_object(note)


@bot.startup
async def startup(poster: MarkdownPoster):
    await poster("__Booo!__ 🐦")



from typing import Callable
from dataclasses import dataclass

from roboherd.cow import RoboCow


@dataclass
class HandlerInformation:
    action: str
    activity_type: str
    func: Callable
    cow: RoboCow



import asyncio
import logging

from dataclasses import dataclass, field

from roboherd.cow import RoboCow, CronEntry
from almabtrieb import Almabtrieb

from .manager import HerdManager
from .scheduler import HerdScheduler
from .processor import HerdProcessor

logger = logging.getLogger(__name__)


@dataclass
class RoboHerd:
    name: str = "RoboHerd"
    base_url: str = "http://abel"

    manager: HerdManager = field(default_factory=HerdManager)
    cows: list[RoboCow] = field(default_factory=list)  # type: ignore

    async def run(self, connection: Almabtrieb):
        async with connection:
            self.validate(connection)
            await self.startup(connection)
            await self.process(connection)

    async def check(
        self, connection: Almabtrieb, raise_if_cows_to_create: bool = False
    ):
        async with connection:
            if not connection.information:
                raise Exception("Could not get information from server")
            self.validate(connection)
            cows_to_create = self.manager.cows_to_create(connection.information.actors)

            if len(cows_to_create) > 0:
                print("Missing cows: " + ", ".join([x.name for x in cows_to_create]))

                if raise_if_cows_to_create:
                    raise Exception("Missing cows")

    async def startup(self, connection: Almabtrieb):
        if not connection.information:
            raise Exception("Could not get information from server")

        self.cows = self.manager.existing_cows(connection.information.actors)
        cows_to_create = self.manager.cows_to_create(connection.information.actors)

        for cow_config in cows_to_create:
            logger.info("Creating cow with name %s", cow_config.name)
            cow = cow_config.load()
            result = await connection.create_actor(
                name=f"{self.manager.prefix}{cow_config.name}",
                base_url=cow.internals.base_url or self.base_url,
                preferred_username=cow.information.handle,
                profile={"type": "Service"},
                automatically_accept_followers=True,
            )
            cow.internals.actor_id = result.get("id")

            self.cows.append(cow)

        for cow in self.cows:
            await cow.run_startup(connection)

    async def process(self, connection: Almabtrieb):
        async with asyncio.TaskGroup() as tg:
            logger.info("Starting processing tasks")

            processor = HerdProcessor(connection, self.incoming_handlers())
            processor.create_tasks(tg)

            scheduler = HerdScheduler(self.cron_entries(), connection)
            scheduler.create_task(tg)

            connection.add_on_disconnect(scheduler.stop)

    def validate(self, connection: Almabtrieb):
        result = connection.information
        if result is None:
            raise ValueError("information not retrieved from connection")

        logger.info("Got base urls: %s", ",".join(result.base_urls))

        if self.base_url not in result.base_urls:
            logger.error(
                "Configure base url %s not in base urls %s of server",
                self.base_url,
                ", ".join(result.base_urls),
            )
            raise ValueError("Incorrectly configured base url")

    def cron_entries(self) -> list[tuple[RoboCow, CronEntry]]:
        """Returns the cron entries of all cows"""

        result = []
        for cow in self.cows:
            for cron_entry in cow.internals.cron_entries:
                result.append((cow, cron_entry))

        return result

    def incoming_handlers(self) -> list[RoboCow]:
        result = []
        for cow in self.cows:
            if cow.internals.handlers.has_handlers:
                result.append(cow)
        return result
